/*
 *  $Id: ic_finish.c,v 1.6 1994/01/28 17:31:49 doss Exp $
 *
 *  (C) 1993 by Argonne National Laboratory and Mississipi State University.
 *      All rights reserved.  See COPYRIGHT in top-level directory.
 */

#include <mpi.h>


/*@

MPI_Intercomm_finish - finish the creation of an intercommunicator 

algorithm:
. 1) If I'm the local leader, then wait on the posted sends and receives
     to complete.  Post the receive for the remote group information and
	 wait for it to complete.
. 2) Broadcast information received from the remote leader.  
. 3) Make the inter_communicator from the information we now have.

@*/
int MPI_Intercomm_finish ( inter_request, comm_out )
MPI_Intercomm_request *inter_request;
MPI_Comm              *comm_out;
{
  int local_rank;
  int errno = MPI_SUCCESS;
  MPI_Intercomm_request i_req = (*inter_request);
  MPI_Status status[6];
  MPI_Group remote_group;
  MPI_Comm  new_comm;

  /* Check for invalid arguments */
  if (((inter_request==(MPI_Intercomm_request *)0) && (errno=MPI_ERR_NULL)) ||
	  ((comm_out     ==(MPI_Comm *)0)              && (errno=MPI_ERR_NULL))  )
	return (errno);

  /* If I'm the local leader, finish the information exchange */
  MPI_Comm_rank ( i_req->local_comm, &local_rank );
  if ( local_rank == i_req->local_leader ) {

	/* Wait on the communication requests to finish (all except the */
	/* group information send) */
	MPI_Waitall ( 4, i_req->req, status );

   	/* We now know how big the remote group is, so make it */
	remote_group = MPIR_CreateGroup ( i_req->remote_size );

	/* Post the receive for the group information and wait on the send */
	/* and the receive for the group information */
	MPI_Irecv (remote_group->lrank_to_grank, i_req->remote_size, MPI_INT,
			   i_req->remote_leader, i_req->tag, i_req->peer_comm, 
			   &(i_req->req[5]));
	MPI_Waitall ( 2, &(i_req->req[4]), &(status[4]) );

	/* Now we can broadcast the group information to the other local comm */
	/* members. */
	MPI_Bcast(&(i_req->remote_size),1,MPI_INT,local_rank,i_req->local_comm);
	MPI_Bcast(remote_group->lrank_to_grank, i_req->remote_size, MPI_INT,
			  local_rank, i_req->local_comm);
  }
  /* Else I'm just an ordinary comm member, so receive the bcast'd */
  /* info about the remote group */
  else {
	int remote_size;
	MPI_Bcast(&remote_size, 1, MPI_INT, i_req->local_leader,i_req->local_comm);

   	/* We now know how big the remote group is, so make it */
	remote_group = MPIR_CreateGroup ( remote_size );
	
	/* Receive the group info */
	MPI_Bcast(remote_group->lrank_to_grank, remote_size, MPI_INT, 
			  i_req->local_leader, i_req->local_comm);
  }

  /* Broadcast the send context */
  MPI_Bcast(&(i_req->send_context), 1, MPIR_CONTEXT_TYPE, i_req->local_leader, 
			i_req->local_comm);

  /* We all now have all the information necessary, start building the */
  /* inter-communicator */
  new_comm = (*comm_out) = NEW(MPIR_COMMUNICATOR);
  new_comm->comm_type   = MPIR_INTER;
  new_comm->group = remote_group;
  (void) MPIR_GROUP_dup( i_req->local_comm->group, &(new_comm->local_group) );
  new_comm->send_context = i_req->send_context;
  new_comm->recv_context = i_req->recv_context;
  new_comm->comm_cache    = 0;
  new_comm->topology.type = MPI_UNDEFINED;
  (void) MPIR_HBT_new_tree ( &(new_comm->attr_cache) );

  /* Free the intercomm_request structure and set it to null */
  FREE(i_req);
  (*inter_request) = (MPI_Intercomm_request)0;

  return (errno);
}
